/*
  META TAG GENERATOR
  Creates or updates meta tags in your pages
  Works on a local image of the site.
  A free open-source scriptol program by D.G. Sureau
  http://www.scriptol.net
  Licence: OSS (free to use and distribute)

                                    RULES
  -----------------------------------------------------------------------------
  keywords: If missing or less than 10, add up to 30-50 words.
            Add words not in the small dict of useless words.
            The weight of words is calculated according to several rules:
            - already among keywords, always kept
            - present inside the title, heavy weight
            - frequency inside the page
            - and present at top of the page.
  description:
            Kept if already existing, otherwise created from the content of the
            page: the first text that is 150 chars long, or else the longer one.
  robots:   "index, follow" or that is got from metagen.ini
  revisit-after:
            5 days or from metagen.ini
  author:   Webmaster, or the one in metagen.ini
  created:  Kept if existing, otherwise use the date of the html file or that
            in metagen.ini
  revised:  Always the date of the file even if metas unchanged
  language: English, or that in metagen.ini
  -----------------------------------------------------------------------------
*/

include "path.sol"
include "strtools.sol"
include "libphp.sol"

void usage()
  print
  print "Meta Tag Generator v.1.6 - www.scriptol.net"
  print "-=".dup(32) + "-"
  print "Creating or updating meta-tags inside html files."
  print "Parsing the page and recursively any locally linked page."
  print "-".dup(65)
  print "Usage:   metagen [options] page"
  print "Example: metagen c:\\site\\index.html"
  print "Options:"
  print " -q        quiet, don't display changes"
  print " -r1..n    depth of links (-r to parse the page only)"
  print " -f        force replacement for all meta-tags"
  print " -v        display results only, don't change pages"
  print " -g        output to a log file rather than on screen"
  exit(1)
return

dict model             ` default values for meta-tags

text pageText
text basePath
text baseHost = ""     ` base path in links
boolean DISPLAY        ` displaying if true
boolean LOGFLAG        ` using log file, no display
array parsedList = ()  ` list of parsed pages
array levelList = ()
array pageList = ()    ` list of pages, created from links
array dynList = ()     ` list of pages to process

array logfile = ()
text logname           ` path and name of the log file

void display(text t)
   if LOGFLAG
      logfile.push(t + "\n")
   else
      print t
   /if
return


` compare urls, ignore slash/antislash difference

boolean urlCompare(text p1, text p2)
  if p1.length() <> p2.length() return false
  for int i in 0 -- p1.length()
    if p1[i] = "\\" continue
    if p1[i] = "/"  continue
    if p1[i] <> p2[i] return false
  /for
return true

` change antislashes for slashes

text makeUniform(text p)
  p = p.trim()
  for int i in 0 -- p.length()
    if p[i] = "\\"
      p[i] = "/"
      continue
    /if
  /for
return p


` test if the address is included into

boolean hasHost(text theurl)
   if "http://" in theurl ? return true
   if "ftp://"  in theurl ? return true
return false


` get the internet address

text getHost(text theurl)
   boolean f1 = false
   int offset = 0
   offset = theurl.find("http://")
   if offset = nil
      offset = theurl.find("ftp://")
      f1 = true
   /if
   if offset = nil ? return ""
   offset + 6
   if not f1 ? offset + 1

   text sub = theurl[offset..]
   offset = sub.find("/", offset)
   if offset = nil ? return theurl       ` no file or subdir
return theurl[--offset]


` get the part of tag inside quotes or double-quotes, thus the url

text extractLink(int off)
   text c
   text link = ""
   while off < pageText.length()
     c = pageText[off]
     if c = '='  ? continue
     if c = ' '  ? continue
     if c = '"'  : off + 1; break; /if
     if c = '\'' : off + 1; break; /if
     break
   let off + 1
   while off < pageText.length()
     c = pageText[off]
     if c in ('\'', '"', '>')  ? break
     link + c             ` anything else is part of the url
   let off + 1
return link.lower().trim()

` Get the URL from the text, skipping other chars

void getLinks(constant alias text localpath, constant alias text fname, int level)
  array x = ()
  x.load(basePath + fname)
  //print "getlink in " , basePath + fname
  array pageLinks = ()
  ` removing ending special codes
  scan x let x[] = x[].toText().rtrim().lower()
  pageText = x.join("")            ` merging lines into a text
  int offset = 0
  int srcoff = 0
  while forever
    srcoff = offset
    offset = pageText.find("href=", offset)
    if offset = nil ? offset = pageText.find("src=", srcoff)
    if offset = nil ? break
    offset + 4
    text link = extractLink(offset)
    if link[0] = '#' continue
    if link    = ""  continue
    if hasHost(link) continue
    /*
    if hasHost(link)
       text host = getHost(link)
       if host != baseHost
          continue
       /if
    /if
    */
    link = localpath + link
    if makeUniform(link) in pageList  continue
    dynList.push(link)
    pageList.push(makeUniform(link))
    levelList.push(level)
  /while
return


` read an ".ini" file, parse lines and return a dict of options

dict readIni(text inipath)

  array ini
  inipath = Path.changeExt(inipath, "ini")
  if not file_exists(inipath) return ()
  ini.load(inipath)
  if(ini.size() = 0)
    display(inipath +" file not loaded")
    return ()
  /if
  dict d = ()

  for text option in ini
    option = option.trim()        ` remove trailing spaces and eol
    int i = option.find(":")      ` parse a line
    if i = nil:
      i = option.find("=")
      if i = nil ? continue       ` no valid couple key:value
    /if
    text k = option[--i].trim()   ` key
    text v = option[i+1..].trim() ` value
    if k
    = "logfile": logname = v
    else
      d[k] = v                      ` put them in dict
    /if
  /for

return d



` The META TAGS CLASS
` dedicated for generating the tags

class MetaTags

  array page                   ` content of a web page
  array newpage
  text pageName                ` name of the page
  static boolean FORCE = false ` replace meta tags anyway if true
  static boolean VIEWONLY = false ` don't replace

  ` these variables contains default values for meta tags
  static dict model          ` defaults are common to all pages, thus static
  static text model_author   = "webmaster"
  static text model_robots   = "index, follow"
  static text model_language = "english"
  static text model_created  = ""
  static text model_revisit  = "5-days"

  text keywords            ` list of keywords
  dict  words              ` main list of words with number of occurence
  dict  testwords          ` list of words from the page
  array tempwords          ` temporary list of words
  text description         ` the description meta tag
  text revised             ` the revision-date meta
  static array smalldict   ` list of word not useful as keywords
  static text dictname = "small.en"


  array internalTags = ("<b>", "<i>", "<u>", "<s>", "</b>",
     "</i>", "</s>", "</u>", "<br>" )


  dict symDict = ("nbsp" : " ", "eacute": "", "egrave": "",
                  "ccedil": "", "ocirc" : "", "ecirc" : "",
                  "acirc": "", "icirc":"", "agrave": "",
                  "iuml": "", "uuml": "", "ugrave": "",
                  "lt" : "<",    "gt" : ">" )


  ` constructor

  void MetaTags(text fname)
    pageName = fname
  return

  ` get the list of words not to include as keywords

  static void loadDict(text inipath)
    text ext
    text lang = model_language
    if lang.length() > 1
       lang = lang.trim().lower()
       ext = "." + lang[ .. 1]      ` en for english
    /if
    if ext = nil let ext = ".en"
    dictname = "small" + ext
    text a,b
    a,b = Path.splitFile(inipath)
    dictname = Path.merge(a, dictname)
    smalldict.load(dictname)
    scan smalldict let smalldict[] = smalldict[].trim()
  return

  ` default items are kept from ini file, if exist here  else
  ` they keep their already assigned value (thanks to conditionnal assignment)

  static void setModel(dict m)
   m.display()
    model = m
    model_author   := model["author"]
    model_robots   := model["robots"]
    model_language := model["language"]
    model_created  := model["created"]
    model_revisit  := model["revisit"]
  return

  void setPage(array a) : page = a; return




  ` startScript and endScript are not used for now

  boolean startScript(text line, text c, int pos)
    if c = "<"
      if line[pos + 1] in ("?", "%") return true
      if (pos + 6) < line.length()
         if line[pos + 1 .. pos + 6].lower() = "script" return true
      /if
    /if
  return false

  boolean endScript(text line, text c, int pos)
    if c in ("?", "%")
      if line[pos + 1] = ">" return true
    /if
    if c = "/"
      if (pos + 6) < line.length()
         if line[pos + 1 .. pos + 6].lower() = "script" return true
      /if
    /if
  return false

  ` extracts a 150-255 chars description

  void makeDesc()
    boolean tag = false      ` inside a tag
    boolean body = false     ` inside body
    boolean bodytag = false  ` inside body tag
    boolean inscript = false ` inside script
    boolean symbol           ` special things as &nbsp;
    text    symbolStr        ` the symbol itself
    int pos                  ` current pos in line
    int bodypos = 0
    text subline             ` part of the line

    ` search for the body tag
    text best = ""          ` preferred text
    text desc = ""          ` current text

    for text line in page

      text lowered = line.lower()

      if body

        ` complete the body tag if not closed yet

        if bodytag
           if not (line.find(">") = nil) let bodytag = false
           continue
        /if

        if "/body" in lowered break     ` end of body, exit
        if not inscript
          if "<script" in lowered let inscript = true
          if "<?" in line let inscript = true
          if "<%" in line let inscript = true
        /if
        if inscript
          if "/script" in lowered let inscript = false
          if "?>" in line let inscript = false
          if "%>" in line let inscript = false
          continue
        /if
        symbol = false
        symbolStr = ""
        pos = -1
        int l = line.length()

        ` scan the current line
        for text c in line
           pos + 1

           ` inside tag, continue until the end of tag
           if symbol
              if c = ";"
                 symbol = false
                 c = symDict[symbolStr]    ` c will hold replacing char
                 if c in ("", "false") let c = " "
              else
                 symbolStr + c
                 continue
              /if
           /if
           /*
           if inscript
             if endScript(line, c, pos) let inscript = false
             continue
           /if
           */
           if tag
             if c = ">" let tag = false
             //echo c
             continue
           /if
           /*
           if not inscript
             if startScript(line, c, pos)
                inscript = true
                continue
             /if
           /if
           */
           if not tag
             if c = "<"      ` end of text
               ` removing start or end of bold tag, italic  and so ones...
               ` I leave the last char as it will be removed at end of loop
               if pos < (l - 4)
                  if line[pos .. pos + 3].lower() in internalTags
                     //print "subline=", line[pos .. pos + 3]
                     line[pos .. pos + 2] = ""
                     continue
                  /if
               /if
               if pos < (l - 3)
                  if line[pos .. pos + 2].lower() in internalTags
                     //print "subline=", line[pos .. pos + 2], line
                     line[pos .. pos + 1] = ""
                     continue
                  /if
               /if
               tag = true
               if desc <> nil
                 if desc.length() > 150  ` reached what we want
                    ` the description must be shortened to 255 chars or less
                    int dlen = desc.length()
                    //print "desc=", dlen, desc
                    if dlen > 254
                      dlen = 254
                      while (desc[dlen] = " ") and (dlen > 140)
                        dlen - 1
                      /while
                    /if
                    description = desc[ -- dlen].trim()
                    //print "desc=", description
                    return
                 /if
                 if desc.length() > best.length() let best = desc
                 desc = ""
               /if
               continue
             /if
             if c = "&"
               symbol = true
               symbolStr = ""
               continue
             /if
             //echo c
             desc + c
           /if
        /for
      else      ` not inside body
        bodypos = lowered.find("<body")
        if not (bodypos = nil)
           body = true
           if line.find(">", bodypos) = nil let bodytag = true
        /if
      /if
    /for
    //print "desc=", desc
    //print "best=", best
    if desc.length() > best.length() let best = desc
    description = best.trim()   ` now assign the longer text found
  return


  ` tests if the word is a valid keyword
  ` you may change the rules here

  boolean validKeyword(text w)
     if w in smalldict return false
     if int(w) <> 0 return false
     if w.length() < 2 return false
     if w[0] = "#" return false
     if w[0] = "%" return false
     if w in StrTools.cdelimiters return false
  return true


  ` add a word to the final list with the proper number of occurences

  void addToWords(text w, int value)
    w = w.trim().lower()
    if words[w] <> nil
       words[w] + value
    else
       if validKeyword(w) let words[w] = value
    /if
  return

  ` add a word to the test list

  void addToTest(text w)
     w = w.trim().lower()
     if testwords[w] <> nil
        testwords[w] + 1
     else
        if validKeyword(w) let testwords[w] = 1
     /if
  return

  static text delimiters = " \\.()[]{},;:?+-*=/&~|\"\'<>!\r\n\t@"

  void makeWords()
    boolean tag = false      ` inside a tag
    boolean body = false     ` inside body
    boolean bodytag = false  ` inside body tag
    boolean inscript = false ` inside script
    boolean symbol           ` special things as &nbsp;
    text    symbolStr        ` the symbol itself
    text keyline = ""        ` current text
    int pos
    int bodypos = 0


    testwords = ()

    ` search for body

    for text line in page

      text lowered = line.lower()

      if body

        if bodytag
           if not (line.find(">") = nil) let bodytag = false
           continue
        /if

        if "/body" in lowered break     ` end of body, exit
        if not inscript
          do
          case "<script" in line : inscript = true
          case "<?" in line      : inscript = true
          case "<%" in line      : inscript = true
          /do
        /if
        if inscript
          define match as "strstr"
          if line
          match "/script" : inscript = false
          match "?>"      : inscript = false
          match "%>"      : inscript = false
          /if
          continue
        /if

        symbol = false
        symbolStr = ""
        pos = -1

        ` scan the current keyline
        for text c in line
           pos + 1

           ` inside tag, continue until the end of tag
           if symbol
              if c = ";"
                 symbol = false
                 c = symDict[symbolStr]
                 if c in ("", "false") let c = " "
              else
                 symbolStr + c
                 continue
              /if
           /if

           /*
           if inscript
             if endScript(line, c, pos) let inscript = false
             continue
           /if
           */
           if tag
             if c = ">" let tag = false
             continue
           /if
           /*
           if not inscript
             if startScript(line, c, pos)
                inscript = true
                continue
             /if
           /if
           */
           ` outside tag, concatenate chars of text
           if not tag
             if c = "<"      ` end of text
               tag = true

               ` adding word in the keyline to tempwords
               if keyline <> nil
                 tempwords = StrTools.split(keyline, delimiters)
                 scan tempwords let addToTest(tempwords[].trim())
                 keyline = ""
               /if
               continue
             /if
             if c = "&"
               symbol = true
               symbolStr = ""
               continue
             /if
             //echo c
             keyline + c
           /if
        /for
      else      ` not inside body
        bodypos = lowered.find("<body")
        if not (bodypos = nil)
           body = true
           if line.find(">", bodypos) = nil let bodytag = true
        /if
      /if
    /for

    for text k, int v in testwords
      addToWords(k, v)
    /for

  return


  ` get the name of a meta tag

  text getMetaName(text line)

     int a = line.find("<meta")
     if a = nil : display("error") ; return ""; /if
     a = line.find("name", a)
     if a = nil : display("error") ; return ""; /if

     a + 4

     text name = ""
     boolean started = false
     for text c in line[ a .. ]

       if c = " " continue
       if c = "=" continue
       if c in ( '\'', "\"")
         if started break
         started = true
         continue
       /if

       name + c
     /for

  return name


  ` build a meta tag

  text buildMeta(text k, text v)
    text replace = "<meta name=\"" + k + "\" "
    //if k[..3] = "date" let replace + "scheme=\"Day-Month-Year\" "
    replace + "content=\"" + v + "\">\n"
  return replace


  ` add or replace all meta tags

  void rebuild(dict x, dict y)

    boolean INSIDE = false
    boolean BODY = false
    int left, right
    text lowered
    text replace
    int l

    int linenum = 0
    while linenum < page.size()
      text line = page[linenum]
      if BODY
         newpage.push(line)
         continue
      /if
      lowered = line.lower().rtrim()

      if not INSIDE
         left = lowered.find("</head>")
         if not (left = nil)
            //x.display()
            for text k, text v in x
              //print k, "(", v, ")"
              if v <> nil
                 newpage.push(buildMeta(k, v))
                 x[k] = ""
              /if
            /for
            newpage.push(line)
            BODY = true
            continue
         /if

         l = lowered.length()
         left = lowered.find("<meta")
         if not (left = nil)
           ` skipping other kind of metas, only meta followed by name
           text temp = lowered[ left + 5 .. left + 12]
           if "name" not in temp
             newpage.push(line)
             continue
           /if
           INSIDE = true
           text name = getMetaName(lowered)
           ` if this meta is not in our list, keep it
           if x[name] = nil let x[name] = y[name]
           replace = buildMeta(name, x[name])
           x[name] = ""      ` removing the tag
           newpage.push( line[ -- left ] + replace.rtrim() )
           right = lowered.find(">", left)
           if right = nil   ` always inside the tag
             continue
           else             ` tag ended
             line = line[ right + 1 .. ]
             page[linenum] = line
             linenum - 1
             INSIDE = false
           /if
         else
             newpage.push(line)
         /if
      else    ` always inside a meta
        right = lowered.find(">", left)
        if right = nil
           continue        ` line ignored
        else
           line = line[ right + 1.. ]
           INSIDE = false
           page[linenum] = line
           linenum - 1
        /if

      /if
    let linenum + 1

    newpage.store(pageName)
  return


  ` set or change meta-tags

  void updatePage(dict y, boolean DISPLAY)

    ` lowering the keys into x

    dict x
    for text k, text v in y
       x[ k.lower() ] = v
    /for


    ` process description if not found, or force replace required

    if (x["description"] = nil) or FORCE
       makeDesc()
       x["description"] = description.trim()
    /if

    ` process previous keywords

    keywords  = x["keywords"]
    tempwords = StrTools.split(keywords, " ,\r\n\t")
    scan tempwords
      text w = tempwords[].trim().lower()
      if w = "" continue
      words[w] = 100         ` heavy weight for this word
    /scan

    if words.size() < 10

      ` get keywords from description and add to final list
      text tempdesc = x["description"]
      tempwords = StrTools.split(tempdesc, " ,;.:?'\'\"()-_ \n\t\r")
      scan tempwords let addToWords(tempwords[].trim(), 100)

      ` from words and testwords make the final list
      makeWords()                          ` get keywords from page's content
      //testwords.display()
      words.sort()                         ` sort by occurence numbers
      words = words[words.size() - 50 ..]  ` keep best numbers of occurences
      words = array_reverse(words, true)
      //words.display()

      ` make a text from the dict
      array w = words.keys()
      keywords = w.join(", ")

    /if

    x["keywords"] = keywords

    ` keep value from page or use default...

    if (x["author"] = nil)        or FORCE let x["author"] = model_author
    if (x["robots"] = nil)        or FORCE let x["robots"] = model_robots
    if (x["language"] = nil)      or FORCE let x["language"] = model_language
    if (x["revisit_after"] = nil) or FORCE let x["revisit-after"] = model_revisit

    ` process revision date
    revised = date("d-M-Y", filemtime(pageName))
    x["date-revision"] = revised
    if x["date_creation"] = nil let x["date-creation"] = revised

    if DISPLAY or VIEWONLY
      display("")
      display("Now:")
      for text k, text v in x
        if k in ("false", "") continue
        if v in ("false", "") continue
        display(k.upper() + " : " + v)
      /for
    /if

    if not VIEWONLY
      rebuild(x, y)
    /if

    return

/class




` parse a web page for meta tags

int metaPage(text localpath, text fname, int level)
  int counter = 0
  text fullpath

  text node, ext
  node, ext = Path.splitExt(fname)
  ` only html pages are parsed for meta tags
  if ext.lower() not in ("htm", "html", "php", "php3", "php4", "php5", "asp",
    "shtml", "dhtml", "jsp", "sol")  return 0

  fullpath = basePath + localpath + fname   ` make full path
  localpath, node = Path.splitFile(fname)   ` get dir in name

  if file_exists(fullpath)
    display("")
    display(fullpath)
    display("-".dup(fullpath.length()))

    dict x = get_meta_tags(fullpath)
    if DISPLAY
       display("Before:")
       if x.size() > 0
         for text k, text v in x let display(k.upper() + " : " + v)
       else display("*** No any meta tag ***")
       /if
    /if
    parsedList.push(fname.trim())
    counter + 1

    MetaTags meta = MetaTags(fullpath)
    array a
    a.load(fullpath)
    meta.pageName = fullpath
    meta.setPage(a)
    meta.updatePage(x, DISPLAY)

    if level > 0 let  getLinks(localpath, fname, level - 1)
  else
    if fname[..6].lower() not in ("http://", "ftp://")
       display(text("Take note: ") + fname + " not found")
    /if
  /if

return counter



int main(int num, array args)

  text fname, fpath

  logname = Path.changeExt(args[0], "log")

  if num not in 2 .. 6 let usage()

  DISPLAY = true
  LOGFLAG = false
  int RECURSE_LEVEL = 1
  MetaTags.FORCE = false
  MetaTags.setModel(readIni(args[0]))            ` read options
  MetaTags.loadDict(args[0])

  ` options parser

  for text option in args[1..]
    option = option.toText().lower()
    if option[0] in ('-', '/')
       if option[1]
       = "q" : DISPLAY = false
       = "r" : text opt = option[2..]
               RECURSE_LEVEL = int(opt)
       = "f" : MetaTags.FORCE = true      ` static attribute for all pages
       = "v" : MetaTags.VIEWONLY = true
       = "g" : LOGFLAG = true
       /if
    else
      fpath = option
    /if
  /for

  display("Meta Tag Generator 1.4 - www.scriptol.net")
  display("using " + MetaTags.dictname)

  if not Path.exists(fpath)
    display(fpath + " not found")
    if LOGFLAG let logfile.store(logname)
    exit()
  /if

  int counter = 0
  basePath, fname = Path.splitFile(fpath)
  if hasHost(basePath)
     baseHost = getHost(basePath)
     if baseHost.length() = basePath.length() ? fname = fpath
  /if

  ` parsing this page and all further links found during the process

  fname = fname.trim()
  dynList.push(fname)
  pageList.push(makeUniform(fname))
  levelList.push(RECURSE_LEVEL)
  do
     fname = dynList.shift()
     display( "-".dup(10) + " processing " + fname)
     counter  + metaPage("", fname, levelList.shift())
  /do while not dynList.empty()
  display("")

  ` display some infos

  display("--- List of parsed files: " + "-".dup(40))
  scan parsedList let display(parsedList[])
  display("-".dup(65))
  display(text(counter) + " page" + plural(counter))

  if LOGFLAG let logfile.store(logname)

return 0

main($argc, $argv)
